"""
NCP client
==========

.. currentmodule:: ncplib

:mod:`ncplib` allows you to connect to a :doc:`server` and issue commands.


Overview
--------

Connecting to a NCP server
^^^^^^^^^^^^^^^^^^^^^^^^^^

Connect to a :doc:`server` using :func:`connect`. The returned :class:`Connection` will automatically close when the
connection block exits.

.. code:: python

    import ncplib

    async with await ncplib.connect("127.0.0.1", 9999) as connection:
        pass  # Your client code here.

    # Connection is automatically closed here.


Sending a packet
^^^^^^^^^^^^^^^^

Send a :term:`NCP packet` containing a single :term:`NCP field` using :meth:`Connection.send`:

.. code:: python

    response = connection.send("DSPC", "TIME", SAMP=1024, FCTR=1200)


Receiving replies to a packet
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The return value of :meth:`Connection.send` is a :class:`Response`. Receive a single :term:`NCP field` reply using
:meth:`Response.recv`:

.. code:: python

    field = await response.recv()

Alternatively, use the :class:`Response` as an *async iterator* to loop over multiple :term:`NCP field` replies:

.. code:: python

    async for field in response:
        pass

.. important::
    The *async for loop* will only terminate when the underlying connection closes.


Accessing field data
^^^^^^^^^^^^^^^^^^^^

The return value of :meth:`Response.recv` is a :class:`Field`, representing a :term:`NCP field`. Access contained
:term:`NCP parameters <NCP parameter>` using item access:

.. code:: python

    print(field["TSDC"])


Advanced usage
^^^^^^^^^^^^^^

-   :doc:`NCP connection documentation <connection>`.


API reference
-------------

.. autofunction:: connect
"""
from __future__ import annotations
import asyncio
from functools import partial
import logging
import platform
from typing import Optional
import warnings
from ncplib.connection import DEFAULT_TIMEOUT, _wait_for, Connection, Field
from ncplib.errors import CommandError, CommandWarning


__all__ = (
    "connect",
)


logger = logging.getLogger(__name__)


def _client_predicate(field: Field, *, auto_erro: bool, auto_warn: bool, auto_ackn: bool) -> bool:
    if auto_erro:
        error_detail = field.get("ERRO")
        error_code = field.get("ERRC")
        if error_detail is not None or error_code is not None:
            raise CommandError(field, error_detail, error_code)  # type: ignore
        # Ignore the rest of packet-level errors.
        if field.name == "ERRO":  # pragma: no cover
            return False
    # Handle warnings.
    if auto_warn:
        warning_detail = field.get("WARN")
        warning_code = field.get("WARC")
        if warning_detail is not None or warning_code is not None:
            warnings.warn(CommandWarning(field, warning_detail, warning_code))  # type: ignore
        # Ignore the rest of packet-level warnings.
        if field.name == "WARN":  # pragma: no cover
            return False
    # Handle acks.
    return not auto_ackn or "ACKN" not in field


async def connect(
    host: str, port: int = 9999, *,
    remote_hostname: str = None,
    hostname: str = None,
    timeout: Optional[float] = DEFAULT_TIMEOUT,
    auto_link: bool = True,
    auto_auth: bool = True,
    auto_erro: bool = True,
    auto_warn: bool = True,
    auto_ackn: bool = True,
) -> Connection:
    """
    Connects to a :doc:`server`.

    :param str host: The hostname of the :doc:`server`. This can be an IP address or domain name.
    :param int port: The port number of the :doc:`server`.
    :param str remote_hostname: The identifying hostname for the remote end of the connection. If omitted, this will
        be the host:port of the NCP server.
    :param str hostname: The identifying hostname in the client connection. Only applies when ``auto_auth`` is
        enabled. Defaults to the system hostname.
    :param Optional[float] timeout: The network timeout (in seconds). If `None`, no timeout is used, which can lead to
        deadlocks. Applies to: connecting, receiving a packet, closing connection.
    :param bool auto_link: Automatically send periodic LINK packets over the connection.
    :param bool auto_auth: Automatically perform the :term:`NCP` authentication handshake on connect.
    :param bool auto_erro: Automatically raise a :exc:`CommandError` on receiving an ``ERRO`` :term:`NCP parameter`.
    :param bool auto_warn: Automatically issue a :exc:`CommandWarning` on receiving a ``WARN`` :term:`NCP parameter`.
    :param bool auto_ackn: Automatically ignore :term:`NCP fields <NCP field>` containing an ``ACKN``
        :term:`NCP parameter`.
    :raises ncplib.NCPError: if the NCP connection failed.
    :return: The client :class:`Connection`.
    :rtype: Connection
    """
    # Create the network connection.
    reader, writer = await _wait_for(asyncio.open_connection(host, port), timeout)
    connection = Connection(
        reader, writer, partial(_client_predicate, auto_erro=auto_erro, auto_warn=auto_warn, auto_ackn=auto_ackn),
        logger=logger,
        remote_hostname=f"{host}:{port}" if remote_hostname is None else remote_hostname,
        timeout=timeout,
        auto_link=auto_link,
    )
    # Handle auto auth.
    try:
        if auto_auth:
            hostname = hostname or platform.node() or "python3-ncplib"
            # Read the initial LINK HELO packet.
            await connection.recv_field("LINK", "HELO")
            # Send the connection request.
            connection.send("LINK", "CCRE", CIW=hostname)
            # Read the connection response packet.
            await connection.recv_field("LINK", "SCAR")
            # Send the auth request packet.
            connection.send("LINK", "CARE", CAR=hostname)
            # Read the auth response packet.
            await connection.recv_field("LINK", "SCON")
    except BaseException:
        connection.close()
        await connection.wait_closed()
        raise
    # All done!
    connection._start_tasks()
    return connection



"""
NCP server
==========

.. currentmodule:: ncplib

:mod:`ncplib` allows you to create a NCP server and respond to incoming :doc:`client` connections.


Overview
--------

Defining a connection handler
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

A connection handler is a coroutine that starts whenever a new :doc:`client` connects to the server. The provided
:class:`Connection` allows you to receive incoming NCP commands as :class:`Field` instances.

.. code:: python

    async def client_connected(connection):
        pass

When the connection handler exits, the :class:`Connection` will automatically close.


Listening for an incoming packet
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

When writing a :doc:`server`, you most likely want to wait for the connected client to execute a command. Within your
``client_connected`` function, Listen for an incomining :term:`NCP field` using :meth:`Connection.recv`.

.. code:: python

    field = await connection.recv()

Alternatively, use the :class:`Connection` as an *async iterator* to loop over multiple :term:`NCP field` replies:

.. code:: python

    async for field in connection:
        pass

.. important::
    The *async for loop* will only terminate when the underlying connection closes.


Accessing field data
^^^^^^^^^^^^^^^^^^^^

The return value of :meth:`Connection.recv` is a :class:`Field`, representing a :term:`NCP field`.

Access information about the :term:`NCP field` and enclosing :term:`NCP packet`:

.. code:: python

    print(field.packet_type)
    print(field.name)

Access contained :term:`NCP parameters <NCP parameter>` using item access:

.. code:: python

    print(field["FCTR"])


Replying to the incoming field
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Send a reply to an incoming :class:`Field` using :meth:`Field.send`.

.. code:: python

    field.send(ACKN=1)


Putting it all together
^^^^^^^^^^^^^^^^^^^^^^^

A simple ``client_connected`` callback might like this:

.. code:: python

    async def client_connected(connection):
        async for field in connection:
            if field.packet_type == "DSPC" and field.name == "TIME":
                field.send(ACNK=1)
                # Do some more command processing here.
            else:
                field.send(ERRO="Unknown command", ERRC=400)
                break


Start the server
^^^^^^^^^^^^^^^^

Start a new NCP server.

.. code:: python

    loop = asyncio.get_event_loop()
    server = loop.run_until_complete(_start_server(client_connected))
    try:
        loop.run_forever()
    finally:
        server.close()
        loop.run_until_complete(server.wait_closed())


Advanced usage
^^^^^^^^^^^^^^

-   :doc:`NCP connection documentation <connection>`.


API reference
-------------

.. autofunction:: start_server

.. autoclass:: Server
    :members:
"""
from __future__ import annotations
import asyncio
from types import TracebackType
from typing import Awaitable, Callable, Optional, Sequence, Set, Type, TypeVar
import logging
from socket import socket
from ncplib.connection import DEFAULT_TIMEOUT, _wait_for, Connection, Field
from ncplib.errors import NCPError


__all__ = (
    "start_server",
    "Server",
)


T = TypeVar("T")


logger = logging.getLogger(__name__)


def _server_predicate(field: Field) -> bool:
    return True


class Server:

    """
    A :doc:`server`.

    Servers can be used as *async context managers* to automatically shut down the server:

    .. code:: python

        async with server:
            pass

        # Server is automatically shut down.

    .. important::

        Do not instantiate this class directly. Use :func:`start_server` to create a :class:`Server`.
    """

    _client_connected: Callable[[Connection], Awaitable[None]]
    _host: str
    _port: int
    _timeout: Optional[float]
    _auto_link: bool
    _auto_auth: bool
    _handlers: Set[asyncio.Task]

    def __init__(
        self, client_connected: Callable[[Connection], Awaitable[None]], host: str, port: int, *,
        timeout: Optional[float],
        auto_link: bool,
        auto_auth: bool,
    ):
        self._client_connected = client_connected  # type: ignore
        self._host = host
        self._port = port
        # Config.
        self._timeout = timeout
        self._auto_link = auto_link
        self._auto_auth = auto_auth
        # Handlers.
        self._handlers = set()

    async def _run_client_connected(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        connection = Connection(
            reader, writer, _server_predicate,
            logger=logger,
            remote_hostname=":".join(map(str, writer.get_extra_info("peername")[:2])),
            timeout=self._timeout,
            auto_link=self._auto_link,
        )
        try:
            # Handle auto-auth.
            if self._auto_auth:
                connection.send("LINK", "HELO")
                # Read the hostname.
                field = await connection.recv_field("LINK", "CCRE")
                try:
                    connection.remote_hostname = str(field["CIW"])
                except KeyError:
                    # Handle authentication failure.
                    logger.warning("Invalid authentication from %s over NCP", connection.remote_hostname)
                    field.send(ERRO="CIW - This field is required", ERRC=401)
                    return
                # Complete authentication.
                connection.send("LINK", "SCAR")
                await connection.recv_field("LINK", "CARE")
                connection.send("LINK", "SCON")
            # Handle connection.
            connection._start_tasks()
            await self._client_connected(connection)  # type: ignore
        # Close the connection.
        except asyncio.CancelledError:  # pragma: no cover
            raise  # Propagate cancels, not needed in Python3.8+.
        except NCPError as ex:  # Warnings on client decode error.
            logger.warning("Connection error from %s over NCP: %s", connection.remote_hostname, ex)
            if not connection.is_closing():
                connection.send("LINK", "ERRO", ERRO="Bad request", ERRC=400)
        except Exception as ex:
            logger.exception("Unexpected error from %s over NCP", connection.remote_hostname, exc_info=ex)
            if not connection.is_closing():
                connection.send("LINK", "ERRO", ERRO="Server error", ERRC=500)
        finally:
            connection.close()
            try:
                await connection.wait_closed()
            except NCPError as ex:  # pragma: no cover
                logger.warning("Connection error from %s over NCP: %s", connection.remote_hostname, ex)

    def _handle_client_connected(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        handler = asyncio.get_running_loop().create_task(self._run_client_connected(reader, writer))
        handler.add_done_callback(self._handlers.remove)
        self._handlers.add(handler)

    async def _connect(self) -> None:
        self._server = await _wait_for(
            asyncio.start_server(self._handle_client_connected, self._host, self._port),
            self._timeout,
        )
        for s in self.sockets:
            logger.info("Listening on %s:%s over NCP", *s.getsockname()[:2])

    @property
    def sockets(self) -> Sequence[socket]:
        """
        A list of the connected listening sockets.
        """
        return self._server.sockets  # type: ignore

    def close(self) -> None:
        """
        Shuts down the server.

        After calling this method, use :meth:`wait_closed` to wait for the server to fully shut down.

        .. hint::

            If you use the server as an *async context manager*, there's no need to call :meth:`Server.close`
            manually.
        """
        # Close the server.
        self._server.close()
        # Stop handlers.
        for handler in self._handlers:
            handler.cancel()

    async def wait_closed(self) -> None:
        """
        Waits for the server to fully shut down.

        .. important::

            Only call this method after first calling :meth:`close`.

        .. hint::

            If you use the server as an *async context manager*, there's no need to call
            :meth:`Server.wait_closed` manually.
        """
        # Wait for handlers to complete.
        if self._handlers:
            await asyncio.wait(self._handlers)
        # Wait for the server to shut down.
        await _wait_for(self._server.wait_closed(), self._timeout)

    async def __aenter__(self) -> "Server":
        return self

    async def __aexit__(self, exc_type: Optional[Type[T]], exc: Optional[T], tb: Optional[TracebackType]) -> None:
        self.close()
        await self.wait_closed()


async def start_server(
    client_connected: Callable[[Connection], Awaitable[None]],
    host: str = "0.0.0.0", port: int = 9999, *,
    timeout: Optional[float] = DEFAULT_TIMEOUT,
    auto_link: bool = True,
    auto_auth: bool = True,
) -> Server:
    """
    Creates and returns a new :class:`Server` on the given host and port.

    :param client_connected: A coroutine function taking a single :class:`Connection`
            argument representing the client connection. When the connection handler exits, the :class:`Connection`
            will automatically close. If the client closes the connection, the connection handler will exit.
    :param str host: The host to bind the server to.
    :param int port: The port to bind the server to.
    :param Optional[float] timeout: The network timeout (in seconds). If `None`, no timeout is used, which can lead to
        deadlocks. Applies to: creating server, receiving a packet, closing connection, closing server.
    :param bool auto_link: Automatically send periodic LINK packets over the connection.
    :param bool auto_auth: Automatically perform the :term:`NCP` authentication handshake on client connect.
    :return: The created :class:`Server`.
    :rtype: Server
    """
    server = Server(client_connected, host, port, timeout=timeout, auto_link=auto_link, auto_auth=auto_auth)
    await server._connect()
    return server



from __future__ import annotations
from array import array
from datetime import datetime, timezone
from struct import Struct
from typing import Callable, Iterable, List, Tuple, Union
import warnings
from ncplib.errors import DecodeError, DecodeWarning
from ncplib.values import uint


# Packet structs.

PACKET_HEADER_STRUCT = Struct("<4s4sII4sII4s")

FIELD_HEADER_STRUCT = Struct("<4s3s1sI")

PARAM_HEADER_STRUCT = Struct("<4s3sB")


# Sizes.

PACKET_HEADER_SIZE = PACKET_HEADER_STRUCT.size

FIELD_HEADER_SIZE = FIELD_HEADER_STRUCT.size

PARAM_HEADER_SIZE = PARAM_HEADER_STRUCT.size

PACKET_FOOTER_SIZE = 8


# Byte sequences.

PACKET_HEADER = b"\xdd\xcc\xbb\xaa"

PACKET_VERSION = (1).to_bytes(4, "little", signed=False)

PACKET_FOOTER = b"\xaa\xbb\xcc\xdd"

PACKET_FOOTER_NO_CHECKSUM = b"\x00\x00\x00\x00" + PACKET_FOOTER


# Known type codes.

TYPE_I32 = 0x00

TYPE_U32 = 0x01

TYPE_STRING = 0x02

TYPE_RAW = 0x80

TYPE_ARRAY_U8 = 0x81

TYPE_ARRAY_U16 = 0x82

TYPE_ARRAY_U32 = 0x83

TYPE_ARRAY_I8 = 0x84

TYPE_ARRAY_I16 = 0x85

TYPE_ARRAY_I32 = 0x86


# Packet encoding.

ARRAY_TYPE_CODES_TO_TYPE_ID = {
    "B": TYPE_ARRAY_U8,
    "H": TYPE_ARRAY_U16,
    "I": TYPE_ARRAY_U32,
    "b": TYPE_ARRAY_I8,
    "h": TYPE_ARRAY_I16,
    "i": TYPE_ARRAY_I32,
}


Bytes = Union[bytes, bytearray]
Param = Union[Bytes, str, int, uint, bool, array]
Params = Iterable[Tuple[str, Param]]
Fields = Iterable[Tuple[str, int, Params]]
Packet = Tuple[str, int, datetime, bytes, Fields]


def encode_packet(packet_type: str, packet_id: int, timestamp: datetime, info: bytes, fields: Fields) -> bytes:
    timestamp = timestamp.astimezone(timezone.utc)
    # Encode the header.
    packet_header = bytearray(PACKET_HEADER_SIZE)
    PACKET_HEADER_STRUCT.pack_into(
        packet_header, 0,
        PACKET_HEADER,  # Hardcoded packet header.
        packet_type.encode("latin1"),
        0,  # Placeholder for the packet size, which we will calculate soon.
        packet_id,
        PACKET_VERSION,
        int(timestamp.timestamp()), timestamp.microsecond * 1000,
        info,
    )
    chunks: List[Bytes] = [packet_header]
    offset = PACKET_HEADER_SIZE
    # Write the packet fields.
    for field_name, field_id, params in fields:
        field_offset = offset
        # Write the field header.
        field_header = bytearray(FIELD_HEADER_SIZE)
        FIELD_HEADER_STRUCT.pack_into(
            field_header, 0,
            field_name.encode("latin1"),
            b"\x00\x00\x00",  # Placeholder for the field size, which we will calculate soom.
            b"\x00",  # Field type ID is ignored.
            field_id,
        )
        chunks.append(field_header)
        offset += FIELD_HEADER_SIZE
        # Write the params.
        for param_name, param_value in params:
            if isinstance(param_value, uint):
                param_type_id = TYPE_U32
                param_value = param_value.to_bytes(4, "little")
            elif isinstance(param_value, (int, bool)):
                param_type_id = TYPE_I32
                param_value = param_value.to_bytes(4, "little", signed=True)
            elif isinstance(param_value, str):
                param_type_id = TYPE_STRING
                param_value = param_value.encode("utf-8") + b"\x00"
            elif isinstance(param_value, (bytes, bytearray, memoryview)):
                param_type_id = TYPE_RAW
            elif isinstance(param_value, array):
                param_type_id = ARRAY_TYPE_CODES_TO_TYPE_ID[param_value.typecode]
                param_value = param_value.tobytes()
            else:  # pragma: no cover
                raise TypeError(f"Unsupported value type {type(param_value)}")
            # Write the param header.
            param_size = PARAM_HEADER_SIZE + len(param_value)
            param_padding_size = -param_size % 4
            chunks.append(PARAM_HEADER_STRUCT.pack(
                param_name.encode("latin1"),
                ((param_size + param_padding_size) // 4).to_bytes(3, "little"),
                param_type_id,
            ))
            # Write the param value.
            chunks.append(param_value)
            chunks.append(b"\x00" * param_padding_size)
            offset += param_size + param_padding_size
        # Write the field size.
        field_header[4:7] = ((offset - field_offset) // 4).to_bytes(3, "little")[:3]
    # Encode the packet footer.
    chunks.append(PACKET_FOOTER_NO_CHECKSUM)
    # Write the packet size.
    packet_header[8:12] = ((offset + PACKET_FOOTER_SIZE) // 4).to_bytes(4, "little")
    # All done!
    return b"".join(chunks)


# PacketData decoding.

def decode_packet_cps(header_buf: Bytes) -> Tuple[int, Callable[[Bytes], Packet]]:
    (
        packet_header,
        packet_type,
        packet_size,
        packet_id,
        packet_format_id,
        packet_time,
        packet_nanotime,
        packet_info,
    ) = PACKET_HEADER_STRUCT.unpack(header_buf)
    packet_size = packet_size * 4
    if packet_header != PACKET_HEADER:  # pragma: no cover
        raise DecodeError(f"Invalid packet header {packet_header!r}")
    # Decode the rest of the body data.
    size_remaining = packet_size - PACKET_HEADER_SIZE

    def decode_packet_body(buf: Bytes) -> Packet:
        offset = 0
        # Check footer.
        if buf[-4:] != PACKET_FOOTER:  # pragma: no cover
            raise DecodeError(f"Invalid packet footer {buf[-4:]!r}")
        # Decode fields.
        field_limit = size_remaining - PACKET_FOOTER_SIZE
        fields = []
        while offset < field_limit:
            # Decode field header.
            field_name, field_size, field_type_id, field_id = FIELD_HEADER_STRUCT.unpack_from(buf, offset)
            param_limit = offset + int.from_bytes(field_size, "little") * 4
            offset += FIELD_HEADER_SIZE
            # Decode params.
            params = []
            while offset < param_limit:
                # Decode the param header.
                param_name, param_size, param_type_id = PARAM_HEADER_STRUCT.unpack_from(buf, offset)
                param_size = int.from_bytes(param_size, "little") * 4
                # Decode the param value.
                param_value_raw = buf[offset+PARAM_HEADER_SIZE:offset+param_size]
                param_value: Param
                if param_type_id == TYPE_U32:
                    param_value = uint.from_bytes(param_value_raw, "little")
                elif param_type_id == TYPE_I32:
                    param_value = int.from_bytes(param_value_raw, "little", signed=True)
                elif param_type_id == TYPE_STRING:
                    param_value = param_value_raw.split(b"\x00", 1)[0].decode()
                elif param_type_id == TYPE_RAW:
                    param_value = bytes(param_value_raw)
                elif param_type_id == TYPE_ARRAY_U8:
                    param_value = array("B", param_value_raw)
                elif param_type_id == TYPE_ARRAY_U16:
                    param_value = array("H", param_value_raw)
                elif param_type_id == TYPE_ARRAY_U32:
                    param_value = array("I", param_value_raw)
                elif param_type_id == TYPE_ARRAY_I8:
                    param_value = array("b", param_value_raw)
                elif param_type_id == TYPE_ARRAY_I16:
                    param_value = array("h", param_value_raw)
                elif param_type_id == TYPE_ARRAY_I32:
                    param_value = array("i", param_value_raw)
                else:  # pragma: no cover
                    warnings.warn(DecodeWarning("Unsupported type ID", param_type_id))
                # Store the param.
                params.append((param_name.rstrip(b" \x00").decode("latin1"), param_value))
                offset += param_size
                # Check for param overflow.
                if offset > param_limit:  # pragma: no cover
                    raise DecodeError(f"Parameter overflow by {offset - param_limit} bytes")
            # Store the field.
            fields.append((field_name.rstrip(b" \x00").decode("latin1"), field_id, params))
        # Check for field overflow.
        if offset > field_limit:  # pragma: no cover
            raise DecodeError(f"Field overflow by {offset - field_limit} bytes")
        # All done!
        return (
            packet_type.rstrip(b" \x00").decode("latin1"),
            packet_id,
            datetime.fromtimestamp(packet_time, tz=timezone.utc).replace(microsecond=packet_nanotime // 1000),
            packet_info,
            fields,
        )

    # Return the number of bytes to read, and the function to finish decoding.
    return size_remaining, decode_packet_body


def decode_packet(buf: Bytes) -> Packet:
    body_size, decode_packet_body = decode_packet_cps(buf[:PACKET_HEADER_SIZE])
    return decode_packet_body(buf[PACKET_HEADER_SIZE:])  # 32 is the size of the packet header.



"""
ncplib
======

:term:`NCP` library for Python 3, developed by `CRFS`_.

.. image:: https://travis-ci.org/CRFS/python3-ncplib.svg?branch=master
    :target: https://travis-ci.org/CRFS/python3-ncplib


Features
--------

-   :doc:`client`.
-   :doc:`server`.
-   Asynchronous connections via :mod:`asyncio`.


Resources
---------

-   `Documentation`_ is on Read the Docs.
-   `Examples`_, `issue tracking`_ and `source code`_ are on GitHub.
-   `Continuous integration`_ is on Travis CI.


Usage
-----

.. toctree::
    :maxdepth: 1

    installation
    client
    server
    connection
    errors
    values


More information
----------------

.. toctree::
    :maxdepth: 1

    contributing
    glossary
    changelog


.. include:: /_include/links.rst
"""


from ncplib.client import *  # noqa
from ncplib.connection import *  # noqa
from ncplib.errors import *  # noqa
from ncplib.server import *  # noqa
from ncplib.values import *  # noqa



"""
Errors and warnings
===================

.. currentmodule:: ncplib

:term:`NCP` errors and warnings.


API reference
-------------

.. autoexception:: NCPError
    :members:

.. autoexception:: NetworkError
    :members:

.. autoexception:: NetworkTimeoutError
    :members:

.. autoexception:: ConnectionClosed
    :members:

.. autoexception:: CommandError
    :members:

.. autoexception:: DecodeError
    :members:

.. autoexception:: CommandWarning
    :members:

.. autoexception:: DecodeWarning
    :members:
"""
from __future__ import annotations
import asyncio
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ncplib.connection import Field


__all__ = (
    "NCPError",
    "NetworkError",
    "NetworkTimeoutError",
    "ConnectionClosed",
    "CommandError",
    "DecodeError",
    "NCPWarning",
    "CommandWarning",
    "DecodeWarning",
)


class CommandMixin:

    field: Field
    detail: str
    code: int

    def __init__(self, field: Field, detail: str, code: int) -> None:
        super().__init__(f"{field.packet_type} {field.name} {detail!r} (code {code})")  # type: ignore
        self.field = field
        self.detail = detail
        self.code = code


# Errors.

class NCPError(Exception):

    """Base class for all exceptions thrown by :mod:`ncplib`."""


class NetworkError(NCPError, OSError):

    """
    Raised when an NCP :class:`Connection` cannot connect, or disconnects unexpectedly.
    """


class NetworkTimeoutError(NCPError, asyncio.TimeoutError):

    """
    Raised when an NCP :class:`Connection` times out while performing network activity.
    """


class ConnectionClosed(NCPError, EOFError):

    """
    Raised when an NCP :class:`Connection` is closed gracefully.
    """


class CommandError(CommandMixin, NCPError):

    """
    Raised by the :doc:`client` when the :doc:`server` sends a :term:`NCP field` containing an ``ERRO`` parameter.

    Can be disabled by setting ``auto_erro`` to :obj:`False` in :func:`ncplib.connect`.

    .. attribute:: field

        The :class:`ncplib.Field` that triggered the error.

    .. attribute:: detail

        The human-readable :class:`str` message from the server.

    .. attribute:: code

        The :class:`int` code from the server,
    """


class DecodeError(NCPError):

    """
    Raised when a non-recoverable error was encountered in a :term:`NCP packet`.
    """


# Warnings.

class NCPWarning(Warning):

    """Base class for all warnings raised by :mod:`ncplib`."""


class CommandWarning(CommandMixin, NCPWarning):

    """
    Issued by the :doc:`client` when the :doc:`server` sends a :term:`NCP field` containing a ``WARN`` parameter.

    Can be disabled by setting ``auto_warn`` to :obj:`False` in :func:`ncplib.connect`.

    .. attribute:: field

        The :class:`ncplib.Field` that triggered the error.

    .. attribute:: detail

        The human-readable :class:`str` message from the server.

    .. attribute:: code

        The :class:`int` code from the server,
    """


class DecodeWarning(NCPWarning):

    """
    Issued when a recoverable error was encountered in a :term:`NCP packet`.
    """



"""
Value types
===========

.. currentmodule:: ncplib

Overview
--------

NCP data types are mapped onto python types as follows:

=========== ================================================
NCP type    Python type
=========== ================================================
int32       :class:`int`
uint32      :class:`ncplib.uint`
string      :class:`str`
raw         :class:`bytes`
data int8   :class:`array.array(typecode="b") <array.array>`
data int16  :class:`array.array(typecode="h") <array.array>`
data int32  :class:`array.array(typecode="i") <array.array>`
data uint8  :class:`array.array(typecode="B") <array.array>`
data uint16 :class:`array.array(typecode="H") <array.array>`
data uint32 :class:`array.array(typecode="I") <array.array>`
=========== ================================================


API reference
-------------

.. autoclass:: uint
    :members:
"""
from __future__ import annotations

__all__ = (
    "uint",
)


# Type converters.

class uint(int):

    """
    An unsigned integer value.

    Python does not distinguish between signed and unsigned integers, but :term:`NCP` encodes them differently.
    Additionally, the binary encoding restricts the ranges of signed and unsigned integers:

    -   ``-2 ** 31 <= int32 <= 2 ** 31 - 1``
    -   ``0 <= uint32 <= 2 ** 32 - 1``

    To distinguish between ``int32`` and ``uint32`` in :term:`NCP parameter` values, wrap any :class:`int` values to be
    encoded as ``uint32`` in :class:`uint`.
    """

    __slots__ = ()



from setuptools import setup, find_packages


setup(
    name="ncplib",
    version="4.1.1",
    license="BSD",
    description="CRFS NCP library for Python 3.",
    author="Dave Hall",
    author_email="dhall@crfs.com",
    url="https://github.com/CRFS/python3-ncplib",
    packages=find_packages(exclude=["tests", "examples"]),
    package_data={"ncplib": ["py.typed"]},
    install_requires=[
        "async_timeout>=3.0,<4.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



"""
NCP connection
==============

.. currentmodule:: ncplib

:term:`NCP` connections are used by the :doc:`client` and :doc:`server` to represent each side of a connection.

Overview
--------

Getting started
^^^^^^^^^^^^^^^

-   :doc:`NCP client documentation <client>`.
-   :doc:`NCP server documentation <client>`.


Spawning tasks
^^^^^^^^^^^^^^

Spawn a concurrent task to handle long-running commands:

.. code::

    import asyncio

    loop = asyncio.get_event_loop()

    async def handle_dspc_time(field):
        field.send(ACKN=1)
        await asyncio.sleep(10)  # Simulate a blocking task.
        field.send(TSDC=0, TIMM=1)

    for field in connection:
        if field.packet_type == "DSPC" and field.name == "TIME":
            # Spawn a concurrent task to avoid blocking the accept loop.
            loop.create_task(handle_dspc_time(field))
        # Handle other field types here.


API reference
-------------

.. important::

    Do not instantiate these classes directly. Use :func:`connect` to create a :doc:`client` connection.
    Use :func:`start_server` to create a :doc:`server`.

.. autoclass:: Connection
    :members:

.. autoclass:: Response
    :members:

.. autoclass:: Field
    :members:
"""
from __future__ import annotations
import asyncio
from async_timeout import timeout
from datetime import datetime, timezone
from itertools import cycle
import logging
from time import time
from types import TracebackType
from typing import AsyncIterator, Awaitable, Callable, Dict, List, Mapping, Optional, Set, Tuple, Type, TypeVar
from uuid import getnode as get_mac
from ncplib.errors import NetworkError, NetworkTimeoutError, ConnectionClosed, DecodeError
from ncplib.packets import Packet, Param, Params, Fields, encode_packet, decode_packet_cps, PACKET_HEADER_SIZE


__all__ = (
    "Connection",
    "Response",
    "Field",
)


T = TypeVar("T")


# The last four bytes of the MAC address is used as an ID field.
CLIENT_ID = get_mac().to_bytes(6, "little")[-4:]

# The footer and other metadata for the LINK heartbeat packet trailer.
LINK_TRAILER = b"".join((b'\x00\x00\x00\x00', CLIENT_ID, b'\x00\x00\x00\x00\xaa\xbb\xcc\xdd'))

# ID generation.
_gen_id = cycle(range(2 ** 32)).__next__


DEFAULT_TIMEOUT = 15


async def _wait_for(coro: Awaitable[T], ms: Optional[float]) -> T:
    try:
        async with timeout(ms):
            return await coro
    except asyncio.CancelledError:  # pragma: no cover
        raise  # Propagate cancels, not needed in Python3.8+.
    except asyncio.TimeoutError as ex:  # pragma: no cover
        raise NetworkTimeoutError(ex) from ex
    except OSError as ex:  # pragma: no cover
        raise NetworkError(ex) from ex


class Field(Dict[str, Param]):

    """
    A :term:`NCP field` received by a :class:`Connection`.

    Access :term:`NCP parameter` values using item access:

    .. code:: python

        print(field["PDAT"])

    .. attribute:: connection

        The :class:`Connection` that created this field.

    .. attribute:: packet_type

        The type of :term:`NCP packet` that contained this field. This will be a valid :term:`identifier`.

    .. attribute:: packet_id

        The ID of the of :term:`NCP packet` that contained this field.

    .. attribute:: packet_timestamp

        A timezone-aware :class:`datetime.datetime` describing when the containing packet was sent.

    .. attribute:: name

        The name of the :term:`NCP field`. This will be a valid :term:`identifier`.

    .. attribute:: id

        The unique :class:`int` ID of this field.
    """

    __slots__ = ("connection", "packet_type", "packet_id", "packet_timestamp", "name", "id",)

    connection: Connection
    packet_type: str
    packet_id: int
    packet_timestamp: datetime
    name: str
    id: int

    def __init__(
        self, connection: Connection,
        packet_type: str, packet_id: int, packet_timestamp: datetime,
        name: str, id: int, params: Params,
    ) -> None:
        super().__init__(params)
        self.connection = connection
        self.packet_type = packet_type
        self.packet_id = packet_id
        self.packet_timestamp = packet_timestamp
        self.name = name
        self.id = id

    def __repr__(self) -> str:  # pragma: no cover
        return f"<Field {self.packet_type!r} {self.name!r} {dict(self.items())!r}>"

    def send(self, **params: Param) -> Response:
        """
        Sends a :term:`NCP packet` containing a single field in reply to this field.

        :param \\**params: Keyword arguments, one per :term:`NCP parameter`. Each parameter name should be a valid
            :term:`identifier`, and each parameter value should be one of the supported
            :doc:`value types <values>`.
        :return: A :class:`Response` providing access to any :class:`Field` instances received in reply to
            the sent packet.
        :rtype: Response
        :raises ValueError: if any of the packet, field or parameter names were not a valid :term:`identifier`, or any
            of the parameter values were invalid.
        :raises TypeError: if any of the parameter values were not one of the supported
            :doc:`value types <values>`.
        """
        return self.connection._send_packet(self.packet_type, self.packet_id, [(self.name, self.id, params.items())])


class AsyncIteratorMixin:

    __slots__ = ()

    def __aiter__(self) -> AsyncIterator[Field]:
        return self

    async def __anext__(self) -> Field:
        try:
            return await self.recv()  # type: ignore
        except ConnectionClosed:
            raise StopAsyncIteration


class Response(AsyncIteratorMixin):

    """
    A response to a :term:`NCP packet`, returned by :meth:`Connection.send`, :meth:`Connection.send_packet` and
    :meth:`Field.send`.

    Provides access to any :class:`Field` received in reply to the sent packet.

    Responses can be used as *async iterators* to loop over each incoming :class:`Field`:

    .. code:: python

        async for field in response:
            pass

    .. important::
        The *async for loop* will only terminate when the underlying connection closes.
    """

    __slots__ = ("_connection", "_packet_type", "_packet_id", "_expected_fields")

    _connection: Connection
    _packet_type: str
    _packet_id: int
    _expected_fields: Set[Tuple[str, int]]

    def __init__(
        self, connection: Connection,
        packet_type: str, packet_id: int,
        expected_fields: Set[Tuple[str, int]],
    ) -> None:
        self._connection = connection
        self._packet_type = packet_type
        self._packet_id = packet_id
        self._expected_fields = expected_fields

    async def recv(self) -> Field:
        """
        Waits for the next :class:`Field` received in reply to the sent :term:`NCP packet`.

        :raises ncplib.NCPError: if a field could not be retrieved from the connection.
        :return: The next :class:`Field` received.
        :rtype: Field
        """
        while True:
            field = await self._connection.recv()
            if (
                field.packet_type == self._packet_type and
                field.packet_id == self._packet_id and
                (field.name, field.id) in self._expected_fields
            ):
                return field

    async def recv_field(self, field_name: str) -> Field:
        """
        Waits for the next matching :class:`Field` received in reply to the sent :term:`NCP packet`.

        .. hint::

            Prefer :meth:`recv` unless the sent packet contained multiple fields.

        :param str field_name: The field name, must be a valid :term:`identifier`.
        :raises ncplib.NCPError: if a field could not be retrieved from the connection.
        :return: The next :class:`Field` received.
        :rtype: Field
        """
        while True:
            field = await self.recv()
            if field.name == field_name:
                return field


class Connection(AsyncIteratorMixin):

    """
    A connection between a :doc:`client` and a :doc:`server`.

    Connections can be used as *async iterators* to loop over each incoming :class:`Field`:

    .. code:: python

        async for field in connection:
            pass

    .. important::
        The *async for loop* will only terminate when the underlying connection closes.

    Connections can also be used as *async context managers* to automatically close the connection:

    .. code:: python

        async with connection:
            pass

        # Connection is automatically closed.

    .. attribute:: logger

        The :class:`logging.Logger` used by this connection. Log messages will be prefixed with the host and port of
        the connection.

    .. attribute:: remote_hostname

        The identifying hostname for the remote end of the connection.

    .. attribute:: timeout

        The network timeout (in seconds). If `None`, no timeout is used, which can lead to deadlocks. Applies to:
        receiving a packet, closing connection.

    """

    logger: logging.Logger
    _reader: asyncio.StreamReader
    _predicate: Callable[[Field], bool]
    _field_buffer: List[Field]
    timeout: Optional[float]
    _writer: asyncio.StreamWriter
    remote_hostname: str
    _auto_link: bool
    _auto_link_task: Optional[asyncio.Task]

    def __init__(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, predicate: Callable[[Field], bool], *,
        logger: logging.Logger,
        remote_hostname: str,
        timeout: Optional[float],
        auto_link: bool,
    ):
        # Logging.
        self.logger = logger
        self.logger.info("Connected to %s over NCP", remote_hostname)
        # Packet reading.
        self._reader = reader
        self._predicate = predicate  # type: ignore
        self._field_buffer = []
        self.timeout = timeout
        # Packet writing.
        self._writer = writer
        # Config.
        self.remote_hostname = remote_hostname
        self._auto_link = auto_link
        self._auto_link_task = None

    @property
    def transport(self) -> asyncio.BaseTransport:
        """
        The :class:`asyncio.WriteTransport` used by this connection.
        """
        return self._writer.transport

    # Background tasks.

    async def _run_auto_link(self) -> None:
        while not self._writer.transport.is_closing():
            self._writer.write(b"".join((
                b'\xdd\xcc\xbb\xaaLINK\n\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00',
                int(time()).to_bytes(4, "little"),
                LINK_TRAILER,
            )))
            await asyncio.sleep(3)

    def _start_tasks(self) -> None:
        if self._auto_link:
            self._auto_link_task = asyncio.get_running_loop().create_task(self._run_auto_link())

    # Receiving fields.

    async def _recv_packet(self) -> Packet:
        # Read the header. If there's no more data in the pipe, it's a graceful close.
        try:
            header_buf = await self._reader.readexactly(PACKET_HEADER_SIZE)
        except asyncio.IncompleteReadError as ex:
            if len(ex.partial) == 0:
                raise ConnectionClosed("Connection closed") from ex
            raise DecodeError(ex) from ex  # pragma: no cover
        # Read the body. This has to be present, or it's an unexpected close.
        size_remaining, decode_packet_body = decode_packet_cps(header_buf)
        try:
            body_buf = await self._reader.readexactly(size_remaining)
        except asyncio.IncompleteReadError as ex:  # pragma: no cover
            raise DecodeError(ex) from ex
        return decode_packet_body(body_buf)

    async def recv(self) -> Field:
        """
        Waits for the next :class:`Field` received by the connection.

        :raises ncplib.NCPError: if a field could not be retrieved from the connection.
        :return: The next :class:`Field` received.
        :rtype: Field
        """
        while True:
            # Return buffered fields.
            if self._field_buffer:
                field = self._field_buffer.pop()
                self.logger.debug(
                    "Received field %s %s from %s over NCP",
                    field.packet_type, field.name, self.remote_hostname
                )
                if self._predicate(field):  # type: ignore
                    return field
            packet_type, packet_id, packet_timestamp, packet_info, fields = await _wait_for(
                self._recv_packet(),
                self.timeout,
            )
            # Store the fields in the field buffer.
            self.logger.debug("Received packet %s from %s over NCP", packet_type, self.remote_hostname)
            self._field_buffer = [
                Field(self, packet_type, packet_id, packet_timestamp, field_name, field_id, params)
                for field_name, field_id, params in fields
            ]
            self._field_buffer.reverse()

    async def recv_field(self, packet_type: str, field_name: str) -> Field:
        """
        Waits for the next matching :class:`Field` received by the connection.

        :param str packet_type: The packet type, must be a valid :term:`identifier`.
        :param str field_name: The field name, must be a valid :term:`identifier`.
        :raises ncplib.NCPError: if a field could not be retrieved from the connection.
        :return: The next :class:`Field` received.
        :rtype: Field
        """
        while True:
            field = await self.recv()
            if field.packet_type == packet_type and field.name == field_name:
                return field

    # Packet writing.

    def _send_packet(self, packet_type: str, packet_id: int, fields: Fields) -> Response:
        encoded_packet = encode_packet(packet_type, packet_id, datetime.now(tz=timezone.utc), CLIENT_ID, fields)
        self._writer.write(encoded_packet)
        self.logger.debug("Sent packet %s to %s over NCP", packet_type, self.remote_hostname)
        expected_fields = set()
        for field_name, field_id, params in fields:
            self.logger.debug("Sent field %s %s to %s over NCP", packet_type, field_name, self.remote_hostname)
            expected_fields.add((field_name, field_id))
        # Create an iterator of response fields.
        return Response(self, packet_type, packet_id, expected_fields)

    # Sending fields.

    def send_packet(self, packet_type: str, **fields: Mapping[str, Param]) -> Response:
        """
        Sends a :term:`NCP packet` containing multiple :term:`NCP fields <NCP field>`.

        .. hint::

            Prefer :meth:`send` unless you need to send multiple fields in a single packet.

        :param str packet_type: The packet type, must be a valid :term:`identifier`.
        :param \\**fields: Keyword arguments, one per field. Each field name should be a valid :term:`identifier`, and
            the field value should be a :class:`dict` of parameter names mapped to parameter values. Each parameter name
            should be a valid :term:`identifier`, and each parameter value should be one of the supported
            :doc:`value types <values>`.
        :return: A :class:`Response` providing access to any :class:`Field` instances received in reply to
            the sent packet.
        :rtype: Response
        :raises ValueError: if any of the packet, field or parameter names were not a valid :term:`identifier`, or any
            of the parameter values were invalid.
        :raises TypeError: if any of the parameter values were not one of the supported
            :doc:`value types <values>`.
        """
        return self._send_packet(packet_type, _gen_id(), [
            (field_name, _gen_id(), field_params.items())
            for field_name, field_params
            in fields.items()
        ])

    def send(self, packet_type: str, field_name: str, **params: Param) -> Response:
        """
        Sends a :term:`NCP packet` containing a single :term:`NCP field`.

        :param str packet_type: The packet type, must be a valid :term:`identifier`.
        :param str field_name: The field name, must be a valid :term:`identifier`.
        :param \\**params: Keyword arguments, one per :term:`NCP parameter`. Each parameter name should be a valid
            :term:`identifier`, and each parameter value should be one of the supported
            :doc:`value types <values>`.
        :return: A :class:`Response` providing access to any :class:`Field` instances received in reply to
            the sent packet.
        :rtype: Response
        :raises ValueError: if any of the packet, field or parameter names were not a valid :term:`identifier`, or any
            of the parameter values were invalid.
        :raises TypeError: if any of the parameter values were not one of the supported
            :doc:`value types <values>`.
        """
        return self._send_packet(packet_type, _gen_id(), [(field_name, _gen_id(), params.items())])

    # Connection lifecycle.

    def is_closing(self) -> bool:
        """
        Returns True if the connection is closing.

        A closing connection should not be written to.
        """
        return self.transport.is_closing()

    def close(self) -> None:
        """
        Closes the connection.

        .. hint::

            If you use the connection as an *async context manager*, there's no need to call :meth:`Connection.close`
            manually.
        """
        # Stop handlers.
        if self._auto_link and self._auto_link_task is not None:
            self._auto_link_task.cancel()
        # Close the connection.
        self._writer.close()
        self.logger.info("Disconnected from %s over NCP", self.remote_hostname)

    async def wait_closed(self) -> None:
        """
        Waits for the connection to finish closing.

        .. hint::

            If you use the connection as an *async context manager*, there's no need to call
            :meth:`Connection.wait_closed` manually.
        """
        await _wait_for(self._writer.wait_closed(), self.timeout)

    async def __aenter__(self: T) -> T:
        return self

    async def __aexit__(self, exc_type: Optional[Type[T]], exc: Optional[T], tb: Optional[TracebackType]) -> None:
        self.close()
        await self.wait_closed()
